package com.taiyoyuden.tyapp_terminal;

import android.Manifest;
import android.app.Activity;
import android.app.AlertDialog;
import android.app.Dialog;
import android.bluetooth.BluetoothAdapter;
import android.bluetooth.BluetoothDevice;
import android.bluetooth.BluetoothGattCharacteristic;
import android.bluetooth.BluetoothGattService;
import android.content.BroadcastReceiver;
import android.content.ComponentName;
import android.content.Context;
import android.content.DialogInterface;
import android.content.IntentFilter;
import android.content.ServiceConnection;
import android.content.pm.PackageManager;
import android.location.Location;
import android.location.LocationManager;
import android.os.Build;
import android.os.Bundle;
import android.os.IBinder;
import android.provider.Settings;
import android.support.v7.app.AppCompatActivity;
import android.text.TextUtils;
import android.util.Log;
import android.view.LayoutInflater;
import android.view.View;
import android.view.ViewGroup;
import android.widget.AdapterView;
import android.widget.BaseAdapter;
import android.widget.Button;
import android.widget.ListView;
import android.widget.TextView;
import android.widget.Toast;
import android.content.Intent;
import com.taiyoyuden.tyapp_terminal.BLEService.LocalBinder;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import static android.content.Context.LOCATION_SERVICE;
import static com.taiyoyuden.tyapp_terminal.Terminal.RX_CHARACTERISTIC_UUID;
import static com.taiyoyuden.tyapp_terminal.Terminal.TX_CHARACTERISTIC_UUID;
import static com.taiyoyuden.tyapp_terminal.Terminal.TY_SERVICE_UUID;

/**
 * Main application activity launched at runtime. This activity connects to the BLEService and initiates
 * scanning for Bluetooth Low Energy compatible devices. Discovered devices will appear as selectable items
 * on the application UI, selection of a device will launch the Terminal activity.
 *
 * Activity will prompt user for permission to access location on first launch of application (necessary
 * due to known Android bug requiring location permission for background BLE scanning). Activity will
 * request for enabling Bluetooth system service if it is disabled. Not accepting this request will exit
 * the application.
 * \date    10/18/2016
 */
public class DeviceSelection extends AppCompatActivity {

    private static final String TAG = "TerminalApp";
    private static final int PERMISSION_REQUEST_FINE_LOCATION = 0;
    private static final int REQUEST_ENABLE_BT = 1;
    private static final int REQUEST_OPEN_TERMINAL = 2;
    private static final int REQUEST_ENABLE_LOCATION = 3;

    /// Status notifying that user declined to enable Bluetooth service \hideinitializer
    public static final String REQUEST_ENABLE_BT_DENIED = "com.taiyoyuden.tyapp_terminal.BLUETOOTH_DENIED";

    private BLEService mService;
    private ArrayList<BluetoothDevice> deviceList;
    private CustomAdapter deviceAdapter;
    private HashMap<String, Integer> RSSI;

    // Build the alert dialog
    AlertDialog.Builder builder;

    /**
     * @brief Initialises UI components, performs a permissions check and binds to BLE service class.
     */
    @Override
    protected void onCreate(Bundle savedInstanceState) {
        super.onCreate(savedInstanceState);

        Log.d(TAG, "onCreate");

        // Initialise layout
        setContentView(com.taiyoyuden.tyapp_terminal.R.layout.activity_main);

        deviceList = new ArrayList<>();
        RSSI = new HashMap<>();
        deviceAdapter = new CustomAdapter(this, deviceList);

        ListView listView = (ListView) findViewById(com.taiyoyuden.tyapp_terminal.R.id.device_list);
        listView.setAdapter(deviceAdapter);
        listView.setOnItemClickListener(listenerItemSelected);

        builder = new AlertDialog.Builder(this);
        builder.setTitle("Location Services Not Active");
        builder.setMessage("Location Services are required for BLE scanning");
        builder.setPositiveButton("OK", new DialogInterface.OnClickListener() {
            public void onClick(DialogInterface dialogInterface, int i) {
            // Show location settings when the user acknowledges the alert dialog
                Intent intent = new Intent(Settings.ACTION_LOCATION_SOURCE_SETTINGS);
                startActivityForResult(intent, REQUEST_ENABLE_LOCATION);
            }
        });
        builder.setNegativeButton("Cancel", new DialogInterface.OnClickListener() {
            @Override
            public void onClick(DialogInterface dialog, int which) {
                dialog.dismiss();
                StartScanning();
            }
        });
        registerReceiver(HardwareStatusReceiver, new IntentFilter(BluetoothAdapter.ACTION_STATE_CHANGED));

        // Check if we have the appropriate permissions for BLE background scanning. If not, wait for user
        // to select an option before we attempt to bind to the service
        if (checkPermissions()) {
            // Bind to BLE service so we can initiate scanning
            Intent intent = new Intent(this, BLEService.class);
            bindService(intent, mServiceConnection, Context.BIND_AUTO_CREATE);
        }

    }

    @Override
    protected void onStart() {
        super.onStart();
        Log.d(TAG, "onStart");
        deviceList.clear();
        deviceAdapter.notifyDataSetChanged();
    }

    ///Checks state of Bluetooth radio and resumes scanning for BLE devices.
    @Override
    protected void onResume() {
        super.onResume();
        Log.d(TAG, "onResume");
        deviceList.clear();
        deviceAdapter.notifyDataSetChanged();

        // Start listening for Bluetooth/Location state updates
        IntentFilter hfilter = new IntentFilter();
        hfilter.addAction(BluetoothAdapter.ACTION_STATE_CHANGED);
        hfilter.addAction(LocationManager.PROVIDERS_CHANGED_ACTION);
        registerReceiver(HardwareStatusReceiver, hfilter);
        registerReceiver(gattStatusReceiver, new IntentFilter(BLEService.ACTION_GATT_SERVICES));
        StartScanning();
    }

    /// Halts BLE scanning while activity is not running in the foreground.
    @Override
    protected void onPause() {
        super.onPause();

        // Stop listening for Bluetooth updates
        try {
            unregisterReceiver(HardwareStatusReceiver);
        } catch (IllegalArgumentException e) {
        }
        try {
            unregisterReceiver(gattStatusReceiver);
        } catch (IllegalArgumentException e) {
        }

        if (mService != null) {
            mService.stopScan(mLeScanCallback);
        }
    }

    @Override
    protected void onDestroy() {
        super.onDestroy();

        try {
            unregisterReceiver(HardwareStatusReceiver);
        } catch (IllegalArgumentException e) {
        }
        try {
            unregisterReceiver(gattStatusReceiver);
        } catch (IllegalArgumentException e) {
        }
        if (mService != null) {
            mService.stopScan(mLeScanCallback);
            unbindService(mServiceConnection);
        }
        Log.d(TAG, "Bye Bye");
    }

    @Override
    protected void onActivityResult(int requestCode, int result, Intent data) {
        switch (requestCode) {
            case REQUEST_ENABLE_BT:
                // We requested to enable the Bluetooth system service
                if (result != RESULT_OK) {
                    Toast.makeText(this, "Application requires Bluetooth", Toast.LENGTH_SHORT).show();
                }
                StartScanning();
                break;
            case REQUEST_OPEN_TERMINAL:
                // Our terminal activity returned

                // Restart scanning
                StartScanning();
                break;
            case REQUEST_ENABLE_LOCATION:
                StartScanning();
                break;
        }
    }

    /**
     * Handles results of permission requests
     *
     * @param requestCode  Value used to request permission dialog
     * @param permissions  List of permissions requested from user
     * @param grantResults Results of each permission request
     */
    @Override
    public void onRequestPermissionsResult(int requestCode, String[] permissions, int[] grantResults) {
        // We only have one permission to deal with
        if (grantResults[0] == PackageManager.PERMISSION_GRANTED) {
            // Start scanning for devices if we bound to the service while waiting for permission
            StartScanning();
        } else {
            // Permission not granted, no point continuing the application. User can relaunch and accept.
            Toast.makeText(this, "Application requires permissions for operation", Toast.LENGTH_SHORT).show();
            finish();
        }
    }

    /**
     * Receiver for BluetoothAdapter state changes (Bluetooth service enabled/disabled)
     */
    private BroadcastReceiver HardwareStatusReceiver = new BroadcastReceiver() {
        @Override
        public void onReceive(Context context, Intent intent) {
            switch (intent.getAction()) {
                // Do a hardware check and start scanning on any hardware state change
                case BluetoothAdapter.ACTION_STATE_CHANGED:
                case LocationManager.PROVIDERS_CHANGED_ACTION:
                    StartScanning();
                    break;
            }
        }
    };

    BroadcastReceiver gattStatusReceiver = new BroadcastReceiver() {
        @Override
        public void onReceive(Context context, Intent intent) {
            String action = intent.getAction();
            Log.d(TAG, action);
            if (action.equals(BLEService.ACTION_GATT_CONNECTED)) {

            }
            if (action.equals(BLEService.ACTION_GATT_DISCONNECTED)) {

            }
            if (action.equals(BLEService.ACTION_GATT_SERVICES)) {
                // Check if the required services are available
                // Check if this is a Taiyo Yuden device
                BluetoothGattService TaiyoYudenService = mService.getService(TY_SERVICE_UUID);
                if (TaiyoYudenService == null) {
                    Log.d(TAG, "Not a TaiyoYuden device");
                    Toast.makeText(getApplicationContext(), "Not a TaiyoYuden device", Toast.LENGTH_SHORT).show();
                    mService.disconnectDevice();
                } else {
                    // Check TX and RX UUIDs
                    BluetoothGattCharacteristic RxChar = TaiyoYudenService.getCharacteristic(RX_CHARACTERISTIC_UUID);
                    BluetoothGattCharacteristic TxChar = TaiyoYudenService.getCharacteristic(TX_CHARACTERISTIC_UUID);
                    if (RxChar == null || TxChar == null) {
                        // TaiyoYuden device doesn't support Tx/Rx communication
                        Log.d(TAG, "No Tx/Rx");
                        Toast.makeText(getApplicationContext(), "Device is not running terminal service", Toast.LENGTH_SHORT).show();
                    } else {
                        Intent terminal = new Intent(getApplicationContext(), Terminal.class);
                        // Stop Bluetooth scanning while the terminal activity is open
                        mService.stopScan(mLeScanCallback);
                        // Start terminal application
                        startActivityForResult(terminal, REQUEST_OPEN_TERMINAL);
                    }
                }
            }

        }
    };

    // Class instance for handling service connection events
    private ServiceConnection mServiceConnection = new ServiceConnection() {

        /**
         *  Service is initialised upon successful connection and scanning for BLE devices is
         *  initiated. Prompts the user to enable Bluetooth system service if disabled.
         */
        public void onServiceConnected(ComponentName className, IBinder rawBinder) {
            mService = ((LocalBinder) rawBinder).getService();

            Log.d(TAG, "onServiceConnected mService= " + mService);

            // Initialise BLE service
            if (!mService.initialiseService()) {
                // There was an error configuring the Bluetooth hardware
                finish();
            } else {
                StartScanning();
            }
        }

        // When we unbind from the service
        public void onServiceDisconnected(ComponentName classname) {
            Log.d(TAG, "onServiceDisconnected");
            mService.stopScan(mLeScanCallback);
            mService = null;
        }
    };

    private void StartScanning(){
        if (mService != null && isBluetoothEnabled() && isLocationEnabled()){
            mService.startScan(mLeScanCallback);
            Log.d(TAG, "Scanning");
        }
    }

    //  Callback interface for receiving device discoveries during scanning
    private BluetoothAdapter.LeScanCallback mLeScanCallback =
            new BluetoothAdapter.LeScanCallback() {

                @Override
                public void onLeScan(final BluetoothDevice device, final int rssi, byte[] scanRecord) {
                    runOnUiThread(new Runnable() {
                        @Override
                        public void run() {
                            if (device.getName() != null) {
                                addDevice(device, rssi);
                            }
                        }
                    });
                }
            };

    // Listener for when a device is selected from the list. Connects to device and initiates service discovery
    private AdapterView.OnItemClickListener listenerItemSelected = new AdapterView.OnItemClickListener() {
        @Override
        public void onItemClick(AdapterView<?> parent, View view, int position, long id) {
            final BluetoothDevice device = deviceList.get(position);
            Log.d(TAG, "Device " + device.getName() + " selected");
            mService.connectToDevice(device.getAddress());
        }
    };

    /**
     * \brief This class extends the BaseAdapter to hold a list of Bluetooth devices discovered during
     * scanning. New devices are allocated a 'device' layout which is populated and added to the list.
     */
    public class CustomAdapter extends BaseAdapter {
        private Context context;
        private List<BluetoothDevice> deviceList;

        CustomAdapter(Context context, List<BluetoothDevice> deviceList) {
            this.context = context;
            this.deviceList = deviceList;
        }

        @Override
        public BluetoothDevice getItem(int position) {
            return this.deviceList.get(position);
        }

        @Override
        public long getItemId(int position) {
            return position;
        }

        @Override
        public int getCount() {
            return deviceList.size();
        }

        @Override
        public View getView(final int position, View convertView, ViewGroup parent) {
            LayoutInflater mInflater = (LayoutInflater) context
                    .getSystemService(Activity.LAYOUT_INFLATER_SERVICE);
            convertView = mInflater.inflate(com.taiyoyuden.tyapp_terminal.R.layout.device, null);

            BluetoothDevice thisDevice = getItem(position);
            Integer rssi = RSSI.get(thisDevice.getAddress());

            ((TextView) convertView.findViewById(com.taiyoyuden.tyapp_terminal.R.id.device_name)).setText(thisDevice.getName());
            ((TextView) convertView.findViewById(com.taiyoyuden.tyapp_terminal.R.id.device_address)).setText(thisDevice.getAddress());
            ((TextView) convertView.findViewById(com.taiyoyuden.tyapp_terminal.R.id.device_rssi)).setText(String.valueOf(rssi) + " dBm");
            return convertView;
        }
    }

    /**
     * @brief Checks if the user has previously accepted permissions request for this application. If not
     * a dialog window is displayed to request permission to access location services.
     * @note Required as of API level 23
     */
    private boolean checkPermissions() {
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
            if (this.checkSelfPermission(Manifest.permission.ACCESS_FINE_LOCATION) != PackageManager.PERMISSION_GRANTED) {
                requestPermissions(new String[]{Manifest.permission.ACCESS_FINE_LOCATION}, PERMISSION_REQUEST_FINE_LOCATION);
                return false;
            }
        }
        return true;
    }

    /**
     * @param device Bluetooth device to add/update
     * @param rssi   Signal strength of device
     * @brief Add a newly discovered device to the list so it can be displayed on the UI. Updates devices
     * already present
     */
    private void addDevice(BluetoothDevice device, int rssi) {
        // Check if device exists in list already
        if (!deviceList.contains(device)) {
            // New device
            deviceList.add(device);
            RSSI.put(device.getAddress(), rssi);       // Single RSSI reading
            deviceAdapter.notifyDataSetChanged();
            Log.d(TAG, "Found " + device.getName());
        }

        //RSSI.put(device.getAddress(),rssi);             // Continuous RSSI updates
        //deviceAdapter.notifyDataSetChanged();
    }

    /**
     * @return True if okay, false if we have to make a request
     * @brief Check state of Bluetooth system service to ensure connection is still valid
     */
    private boolean isBluetoothEnabled() {
        if (!BluetoothAdapter.getDefaultAdapter().isEnabled()) {
            // Request user to enable
            startActivityForResult(new Intent(BluetoothAdapter.ACTION_REQUEST_ENABLE), REQUEST_ENABLE_BT);
            return false;
        }
        Log.d(TAG, "Bluetooth On");
        return true;
    }

    /**
     * @return True if okay, false if we have to make a request
     * @brief Check state of Location system service to allow scanning (req'd in API >=23 - Marshmallow)
     */
    private boolean isLocationEnabled() {
        if (Build.VERSION.SDK_INT >= Build.VERSION_CODES.M) {
            // Get Location Manager and check for GPS & Network location services
            LocationManager lm = (LocationManager) getSystemService(LOCATION_SERVICE);
            boolean gps = lm.isProviderEnabled(LocationManager.GPS_PROVIDER);
            boolean net = lm.isProviderEnabled(LocationManager.NETWORK_PROVIDER);
            if (!gps && !net) {
                Dialog alertDialog = builder.create();
                alertDialog.setCanceledOnTouchOutside(true);
                alertDialog.show();
                return false;
            }
        }
        Log.d(TAG, "Location On");
        return true;
    }

}


